package Kakuro;

import choco.integer.IntDomainVar;
import choco.AbstractProblem;

import java.util.Arrays;
import java.util.List;

/**
 * The Sequence represents a contiguous set of cells constrained
 * by a sumAllDiff in the grid
 **/
public class Sequence {

	/**
	 * The start and end cell of the sequence
	 */
	protected Cell start, end;

	/**
	 * The SumAllDiff problem assiociated with the sequence
	 */
	public SumAlldiffPb spb;

	/**
	 * The representation of the sequence in choco variables referenced
	 * as well in KakuroProblem
	 */
	public IntDomainVar[] x;

	/**
	 * Create a sequence object from a start to an end cell with
	 * the corresponding sumAllDiff constraints.
	 * WARNING: it assumes that start is always above and on the right of end
	 */
	public Sequence(Cell start, Cell end, SumAlldiffPb spb) {
		assert (start.i <= end.i && start.j >= end.j);
		this.start = start;
		this.end = end;
		this.spb = spb;
	}

	/**
	 * Set the x table representing the sequence as
	 * choco variables from the grid given in input
	 * @param matrix
	 */
	public void computeVars(Cell[][] matrix) {
		int i1 = start.i, i2 = end.i;
		int j1 = start.j, j2 = end.j;
		if (i1 == i2) {
			x = new IntDomainVar[j2 - j1 + 1];
			for (int j = j1; j <= j2; j++) {
				x[j - j1] = matrix[i1][j].var;
			}
		} else if (j1 == j2) {
			x = new IntDomainVar[i2 - i1 + 1];
			for (int i = i1; i <= i2; i++) {
				x[i - i1] = matrix[i][j1].var;
			}
		} else {
			throw new Error("the sequence is not consistent");
		}

	}

	/**
	 * Test if this sequence contains s2 (all the domains of the variable
	 * this sequence are containing the domains of the variable of s2) 
	 * @param s2
	 * @return
	 */
	public boolean isMoreGeneral(Sequence s2) {
		return spb.isIncludedIn(s2.x,x);
	}

	/**
	 * add to pb, the constraint as a conjonction of allDiff and sum
	 * @param pb the choco problem to post the sequence
	 */
	public void postNaiveSequence(AbstractProblem pb) {
		pb.post(pb.allDifferent(x));
		pb.post(pb.eq(pb.sum(x), spb.sum));
	}

	/**
	 * add to pb, the constraint as a regular enforcing AC
	 * @param pb the choco problem to post the sequence
	 */
	public void postComplexSum_AllDiff(AbstractProblem pb) {
		if (x.length == 9) {
			pb.post(pb.allDifferent(x));
		} else if (!allInst()){
			pb.post(pb.regular(spb.restrictAndGenerateTuples(x),x));
		}
	}

	/**
	 * Are all cells known in this sequence ?
	 * @return
	 */
	public boolean allInst() {
		for (int i = 0; i < x.length; i++) {
			if(!x[i].isInstantiated()) return false;
		}
		return true;
	}

}
