package Kakuro;

import choco.AbstractProblem;
import choco.Problem;
import choco.ContradictionException;
import choco.util.IntIterator;
import choco.global.regular.DFA;
import choco.integer.IntDomainVar;

import java.util.List;
import java.util.LinkedList;

/**
 * A sum problem representing a sum of variables between 1 and 9
 * that have to be alldifferent.
 * This subproblem is represented as an independant choco problem
 * to provide services on the bounds of each variable of the sequence
 * and the all solutions(tuples) that are compiled as an automaton
 */
public class SumAlldiffPb {

	public int nbvar, sum;

	public AbstractProblem pb;
	public IntDomainVar[] x;
	public int lowerb;
	public int upperb;

	public DFA mostGeneralAutomaton;
	public IntDomainVar[] mostGeneralDomains;

	/**
	 * Create a subproblem of the given size and sum 
	 * @param nbvar
	 * @param sum
	 */
	public SumAlldiffPb(int nbvar, int sum) {
		this.nbvar = nbvar;
		this.sum = sum;
	}

	public void init(int nbvar, int sum) {
		pb.post(pb.boundAllDifferent(x, true));
		pb.post(pb.eq(pb.sum(x), sum));
		pb.worldPush();
	}

	public void computeBounds() {
		pb = new Problem();
		x = pb.makeEnumIntVarArray("x", nbvar, 1, 9);
		init(nbvar, sum);
		computeLB();
		computeUB();
	}

	/**
	 * return an accurate lower bound of each variable of the
	 * subproblem
	 */
	public void computeLB() {
		try {
			if (x.length == 9) {
				lowerb = 1;
			} else {
				int lb = x[0].getInf() - 1;
				boolean consistent = false;
				while (!consistent) {
					lb++;
					pb.worldPush();
					x[0].setVal(lb);
					pb.solve();
					pb.worldPopUntil(0);
					consistent = pb.isFeasible();
				}
				lowerb = lb;
			}
		} catch (ContradictionException e) {
			e.printStackTrace();
			throw new Error("error in get lower bound");
		}
	}

	/**
	 * return an accurate upper bound of each variable of the
	 * subproblem
	 */
	public void computeUB() {
		try {
			if (x.length == 9) {
				upperb = 9;
			} else {
				int ub = x[0].getSup() + 1;
				boolean consistent = false;
				while (!consistent) {
					ub--;
					pb.worldPush();
					x[0].setVal(ub);
					pb.solve();
					pb.worldPopUntil(0);
					consistent = pb.isFeasible();
				}
				upperb = ub;
			}
		} catch (ContradictionException e) {
			e.printStackTrace();
			throw new Error("error in get upper bound");
		}
	}


	public int getUpperBound() {
		return upperb;
	}

	public int getLowerBound() {
		return lowerb;
	}

	/**
	 * If xv0 is included in xv1 then we can reuse the dfa computed for
	 * xv1 if it has already been computed.
	 * @param xv0
	 * @param xv1
	 * @return true is the domain of each variable xv0[i] is included in the
	 * domain of xv1[i].
	 */
	public boolean isIncludedIn(IntDomainVar[] xv0, IntDomainVar[] xv1) {
		if(xv1 == null) return false;
		for (int i = 0; i < xv0.length; i++) {
			if (!isIncludedIn(xv0[i], xv1[i]))
				return false;
		}
		return true;
	}

	public boolean isIncludedIn(IntDomainVar v1, IntDomainVar v2) {
		if (v1.equals(v2)) return true;
		IntIterator it = v1.getDomain().getIterator();
		while (it.hasNext()) {
			int val = it.next();
			if (!v2.canBeInstantiatedTo(val)) return false;
		}
		return v2.getDomainSize() >= v1.getDomainSize();
	}

	/**
	 * Generate the all set of tuples of alldifferent values between 1 and 9
	 * whose sum is equal to sumval
	 *
	 * @param a set of variable whose domains can be used to simplify the problem
	 * @return a dfa encoding all the solutions of the SumAllDiff problems taking the domains
	 * of the sequence giving in argument into account
	 */
	public DFA restrictAndGenerateTuples(IntDomainVar[] xv) {
		if (!isIncludedIn(xv,mostGeneralDomains)) {
			LinkedList<int[]> ltuples = new LinkedList<int[]>();
			int cworld = pb.getWorldIndex();
			pb.worldPush();
			try {
				for (int i = 0; i < nbvar; i++) {
					for (int j = 0; j < 9; j++) {
						if (!xv[i].canBeInstantiatedTo(j + 1))
							x[i].remVal(j + 1);
					}
				}
			} catch (ContradictionException e) {
				System.out.println("ici ?");
				e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
			}
			pb.solve();
			if (pb.isFeasible()) {
				do {
					int[] tps = new int[x.length];
					for (int i = 0; i < x.length; i++) {
						tps[i] = x[i].getVal();
					}
					ltuples.add(tps);
				} while (pb.nextSolution() == Boolean.TRUE);
			}
			pb.worldPopUntil(cworld);
			mostGeneralAutomaton = new DFA(ltuples);
			mostGeneralDomains = xv;
		} 

		return mostGeneralAutomaton;
	}

	public static void main(String[] args) {
		SumAlldiffPb spb = new SumAlldiffPb(3, 23);
		System.out.println("" + spb.getUpperBound());
		System.out.println("" + spb.getLowerBound());
	}

}
